@@ -5,30 +5,35 @@ import socket
 import ssl
 import threading
 from io import BufferedReader, BufferedWriter
-from typing import Callable, Dict, Optional, Tuple, Union
+from typing import Any, Callable, List, Optional, Tuple, Union
 
 import lz4.block
 
-from sqlitecloud.resultset import SQLiteCloudResult, SQLiteCloudResultSet
-from sqlitecloud.types import (
+from sqlitecloud.datatypes import (
     SQLITECLOUD_CMD,
     SQLITECLOUD_DEFAULT,
     SQLITECLOUD_INTERNAL_ERRCODE,
-    SQLITECLOUD_RESULT_TYPE,
     SQLITECLOUD_ROWSET,
     SQLiteCloudConfig,
     SQLiteCloudConnect,
-    SQLiteCloudDataTypes,
-    SQLiteCloudException,
     SQLiteCloudNumber,
     SQLiteCloudRowsetSignature,
     SQLiteCloudValue,
+    SQLiteDataTypes,
+)
+from sqlitecloud.exceptions import (
+    SQLiteCloudException,
+    get_sqlitecloud_error_with_extended_code,
+)
+from sqlitecloud.resultset import (
+    SQLITECLOUD_RESULT_TYPE,
+    SQLiteCloudOperationResult,
+    SQLiteCloudResult,
+    SQLiteCloudResultSet,
 )
 
 
 class Driver:
-    SQLiteCloud_DEFAULT_UPLOAD_SIZE = 512 * 1024
-
     def __init__(self) -> None:
         # Used while parsing chunked rowset
         self._rowset: SQLiteCloudResult = None
@@ -80,38 +85,39 @@ class Driver:
         self, command: str, connection: SQLiteCloudConnect
     ) -> SQLiteCloudResult:
         """
-        Execute a query on the SQLite Cloud server.
+        Execute a command on the SQLite Cloud server.
         """
+        command = self._internal_serialize_command(command)
+
         return self._internal_run_command(connection, command)
 
+    def execute_statement(
+        self,
+        query: str,
+        bindings: Tuple[SQLiteDataTypes],
+        # bindings: Union[Tuple[SQLiteDataTypes], Dict[str, SQLiteDataTypes]],
+        connection: SQLiteCloudConnect,
+    ) -> Union[SQLiteCloudResult, SQLiteCloudOperationResult]:
+        """
+        Execute the statement on the SQLite Cloud server.
+        It supports only the `qmark` style for parameter binding.
+        """
+        command = self._internal_serialize_command(
+            [query] + list(bindings), zero_string=True
+        )
+
+        result = self._internal_run_command(connection, command)
+
+        if result.tag != SQLITECLOUD_RESULT_TYPE.RESULT_ARRAY:
+            return result
+
+        return SQLiteCloudOperationResult(result)
+
     def send_blob(self, blob: bytes, conn: SQLiteCloudConnect) -> SQLiteCloudResult:
         """
         Send a blob to the SQLite Cloud server.
         """
-        try:
-            conn.isblob = True
-            return self._internal_run_command(conn, blob)
-        finally:
-            conn.isblob = False
-
-    def prepare_statement(
-        self,
-        query: str,
-        parameters: Union[
-            Tuple[SQLiteCloudDataTypes], Dict[Union[str, int], SQLiteCloudDataTypes]
-        ],
-    ) -> str:
-        # If parameters is a dictionary, replace the keys in the query with the values
-        if isinstance(parameters, dict):
-            for key, value in parameters.items():
-                query = query.replace(":" + str(key), self.escape_sql_parameter(value))
-
-        # If parameters is a tuple, replace each '?' in the query with a value from the tuple
-        elif isinstance(parameters, tuple):
-            for value in parameters:
-                query = query.replace("?", self.escape_sql_parameter(value), 1)
-
-        return query
+        return self._internal_run_command(conn, self._internal_serialize_command(blob))
 
     def is_connected(
         self, connection: SQLiteCloudConnect, main_socket: bool = True
@@ -130,33 +136,6 @@ class Driver:
 
         return True
 
-    def escape_sql_parameter(self, param):
-        if param is None or param is None:
-            return "NULL"
-
-        if isinstance(param, bool):
-            return "1" if param else "0"
-
-        if isinstance(param, str):
-            # replace single quote with two single quotes
-            param = param.replace("'", "''")
-            return f"'{param}'"
-
-        if isinstance(param, (int, float)):
-            return str(param)
-
-        # serialize buffer as X'...' hex encoded string
-        if isinstance(param, bytes):
-            return f"X'{param.hex()}'"
-
-        if isinstance(param, dict) or isinstance(param, list):
-            # serialize json then escape single quotes
-            json_string = json.dumps(param)
-            json_string = json_string.replace("'", "''")
-            return f"'{json_string}'"
-
-        raise SQLiteCloudException(f"Unsupported parameter type: {type(param)}")
-
     def _internal_connect(
         self, hostname: str, port: int, config: SQLiteCloudConfig
     ) -> socket:
@@ -212,7 +191,9 @@ class Driver:
             connection.config,
         )
 
-        self._internal_run_command(connection, buffer, False)
+        self._internal_run_command(
+            connection, self._internal_serialize_command(buffer.decode()), False
+        )
         thread = threading.Thread(
             target=self._internal_pubsub_thread, args=(connection,)
         )
@@ -262,15 +243,15 @@ class Driver:
                 blen -= nread
                 buffer += data
 
-                SQLiteCloud_number = self._internal_parse_number(buffer)
-                clen = SQLiteCloud_number.value
+                sqlitecloud_number = self._internal_parse_number(buffer)
+                clen = sqlitecloud_number.value
                 if clen == 0:
                     continue
 
                 # check if read is complete
                 # clen is the lenght parsed in the buffer
                 # cstart is the index of the first space
-                cstart = SQLiteCloud_number.cstart
+                cstart = sqlitecloud_number.cstart
                 if clen + cstart != tread:
                     continue
 
@@ -329,7 +310,9 @@ class Driver:
             command = f"UPLOAD DATABASE '{dbname}' {keyarg}{keyvalue}"
 
         # execute command on server side
-        result = self._internal_run_command(connection, command)
+        result = self._internal_run_command(
+            connection, self._internal_serialize_command(command)
+        )
         if not result.data[0]:
             raise SQLiteCloudException(
                 "An error occurred while initializing the upload of the database."
@@ -365,7 +348,9 @@ class Driver:
                     # Upload completed
                     break
         except Exception as e:
-            self._internal_run_command(connection, "UPLOAD ABORT")
+            self._internal_run_command(
+                connection, self._internal_serialize_command("UPLOAD ABORT")
+            )
             raise e
 
     def download_database(
@@ -392,7 +377,10 @@ class Driver:
         """
         exists_cmd = " IF EXISTS" if if_exists else ""
         result = self._internal_run_command(
-            connection, f"DOWNLOAD DATABASE {dbname}{exists_cmd};"
+            connection,
+            self._internal_serialize_command(
+                f"DOWNLOAD DATABASE {dbname}{exists_cmd};"
+            ),
         )
 
         if result.nrows == 0:
@@ -409,7 +397,9 @@ class Driver:
 
         try:
             while progress_size < db_size:
-                result = self._internal_run_command(connection, "DOWNLOAD STEP")
+                result = self._internal_run_command(
+                    connection, self._internal_serialize_command("DOWNLOAD STEP")
+                )
 
                 # res is BLOB, decode it
                 data = result.data[0]
@@ -423,7 +413,9 @@ class Driver:
                 if data_len == 0:
                     break
         except Exception as e:
-            self._internal_run_command(connection, "DOWNLOAD ABORT")
+            self._internal_run_command(
+                connection, self._internal_serialize_command("DOWNLOAD ABORT")
+            )
             raise e
 
     def _internal_config_apply(
@@ -432,51 +424,54 @@ class Driver:
         if config.timeout > 0:
             connection.socket.settimeout(config.timeout)
 
-        buffer = ""
+        command = ""
 
         # it must be executed before authentication command
         if config.non_linearizable:
-            buffer += "SET CLIENT KEY NONLINEARIZABLE TO 1;"
+            command += "SET CLIENT KEY NONLINEARIZABLE TO 1;"
 
         if config.account.apikey:
-            buffer += f"AUTH APIKEY {config.account.apikey};"
+            command += f"AUTH APIKEY {config.account.apikey};"
 
         if config.account.username and config.account.password:
-            command = "HASH" if config.account.password_hashed else "PASSWORD"
-            buffer += f"AUTH USER {config.account.username} {command} {config.account.password};"
+            option = "HASH" if config.account.password_hashed else "PASSWORD"
+            command += f"AUTH USER {config.account.username} {option} {config.account.password};"
 
         if config.account.dbname:
             if config.create and not config.memory:
-                buffer += f"CREATE DATABASE {config.account.dbname} IF NOT EXISTS;"
-            buffer += f"USE DATABASE {config.account.dbname};"
+                command += f"CREATE DATABASE {config.account.dbname} IF NOT EXISTS;"
+            command += f"USE DATABASE {config.account.dbname};"
 
         if config.compression:
-            buffer += "SET CLIENT KEY COMPRESSION TO 1;"
+            command += "SET CLIENT KEY COMPRESSION TO 1;"
 
         if config.zerotext:
-            buffer += "SET CLIENT KEY ZEROTEXT TO 1;"
+            command += "SET CLIENT KEY ZEROTEXT TO 1;"
 
         if config.noblob:
-            buffer += "SET CLIENT KEY NOBLOB TO 1;"
+            command += "SET CLIENT KEY NOBLOB TO 1;"
 
         if config.maxdata:
-            buffer += f"SET CLIENT KEY MAXDATA TO {config.maxdata};"
+            command += f"SET CLIENT KEY MAXDATA TO {config.maxdata};"
 
         if config.maxrows:
-            buffer += f"SET CLIENT KEY MAXROWS TO {config.maxrows};"
+            command += f"SET CLIENT KEY MAXROWS TO {config.maxrows};"
 
         if config.maxrowset:
-            buffer += f"SET CLIENT KEY MAXROWSET TO {config.maxrowset};"
+            command += f"SET CLIENT KEY MAXROWSET TO {config.maxrowset};"
 
-        if len(buffer) > 0:
-            self._internal_run_command(connection, buffer)
+        if len(command) > 0:
+            self._internal_run_command(
+                connection, self._internal_serialize_command(command)
+            )
 
     def _internal_run_command(
         self,
         connection: SQLiteCloudConnect,
-        command: Union[str, bytes],
+        command: bytes,
         main_socket: bool = True,
     ) -> SQLiteCloudResult:
+        """Send serialized command to the server and read the response."""
         if not self.is_connected(connection, main_socket):
             raise SQLiteCloudException(
                 "The connection is closed.",
@@ -489,31 +484,24 @@ class Driver:
     def _internal_socket_write(
         self,
         connection: SQLiteCloudConnect,
-        command: Union[str, bytes],
+        command: bytes,
         main_socket: bool = True,
     ) -> None:
-        # compute header
-        delimit = "$" if connection.isblob else "+"
-        buffer = command.encode() if isinstance(command, str) else command
-        buffer_len = len(buffer)
-        header = f"{delimit}{buffer_len} "
-
-        sock = connection.socket if main_socket else connection.pubsub_socket
-
-        # write header
-        try:
-            sock.sendall(header.encode())
-        except Exception as exc:
-            raise SQLiteCloudException(
-                "An error occurred while writing header data.",
-                SQLITECLOUD_INTERNAL_ERRCODE.NETWORK,
-            ) from exc
+        """
+        Write to the socket the command serialized with the SCSP protocol.
 
+        Args:
+            connection (SQLiteCloudConnect): The connection object to the SQLite Cloud server.
+            command (bytes): The command to send.
+            main_socket (bool): If True, write to the main socket, otherwise write to the pubsub socket.
+        """
         # write buffer
-        if buffer_len == 0:
+        if len(command) == 0:
             return
         try:
-            sock.sendall(buffer)
+            sock = connection.socket if main_socket else connection.pubsub_socket
+
+            sock.sendall(command)
         except Exception as exc:
             raise SQLiteCloudException(
                 "An error occurred while writing data.",
@@ -532,90 +520,115 @@ class Driver:
         slicing the buffer into parts if there are special characters like "ò".
         """
         buffer = b""
-        buffer_size = 8192
+        command_type = ""
+        command_length_value = b""
         nread = 0
 
         sock = connection.socket if main_socket else connection.pubsub_socket
 
+        # read the lenght of the command, eg:
+        # ?LEN <command>, where `?` is any command type
+        # _ for null command
+        # :145 for integer command with value 145
         while True:
             try:
-                data = sock.recv(buffer_size)
+                data = sock.recv(1)
                 if not data:
-                    raise SQLiteCloudException("Incomplete response from server.")
+                    raise SQLiteCloudException(
+                        "Incomplete response from server. Cannot read the command length."
+                    )
             except Exception as exc:
                 raise SQLiteCloudException(
-                    "An error occurred while reading data from the socket.",
+                    "An error occurred while reading command length from the socket.",
                     SQLITECLOUD_INTERNAL_ERRCODE.NETWORK,
                 ) from exc
 
-            # the expected data length to read
-            # matches the string size before decoding it
             nread += len(data)
-            # update buffers
             buffer += data
 
-            c = chr(buffer[0])
+            # first character is the type of the message
+            if nread == 1:
+                command_type = data.decode()
+                continue
 
-            if (
-                c == SQLITECLOUD_CMD.INT.value
-                or c == SQLITECLOUD_CMD.FLOAT.value
-                or c == SQLITECLOUD_CMD.NULL.value
-            ):
-                if not buffer.endswith(b" "):
-                    continue
-            elif c == SQLITECLOUD_CMD.ROWSET_CHUNK.value:
-                isEndOfChunk = buffer.endswith(SQLITECLOUD_ROWSET.CHUNKS_END.value)
-                if not isEndOfChunk:
-                    continue
-            else:
-                SQLiteCloud_number = self._internal_parse_number(buffer)
-                n = SQLiteCloud_number.value
-                cstart = SQLiteCloud_number.cstart
+            # end of len value
+            if data == b" ":
+                break
 
-                can_be_zerolength = (
-                    c == SQLITECLOUD_CMD.BLOB.value or c == SQLITECLOUD_CMD.STRING.value
-                )
-                if n == 0 and not can_be_zerolength:
-                    continue
-                if n + cstart != nread:
-                    continue
+            command_length_value += data
 
+        if (
+            command_type == SQLITECLOUD_CMD.INT.value
+            or command_type == SQLITECLOUD_CMD.FLOAT.value
+            or command_type == SQLITECLOUD_CMD.NULL.value
+        ):
             return self._internal_parse_buffer(connection, buffer, len(buffer))
 
+        command_length = int(command_length_value)
+
+        # read the command
+        nread = 0
+
+        while nread < command_length:
+            buffer_size = min(command_length - nread, 8192)
+
+            try:
+                data = sock.recv(buffer_size)
+                if not data:
+                    raise SQLiteCloudException(
+                        "Incomplete response from server. Cannot read the command."
+                    )
+            except Exception as exc:
+                raise SQLiteCloudException(
+                    "An error occurred while reading the command from the socket.",
+                    SQLITECLOUD_INTERNAL_ERRCODE.NETWORK,
+                ) from exc
+
+            nread += len(data)
+            buffer += data
+
+        return self._internal_parse_buffer(connection, buffer, len(buffer))
+
     def _internal_parse_number(
         self, buffer: bytes, index: int = 1
     ) -> SQLiteCloudNumber:
-        SQLiteCloud_number = SQLiteCloudNumber()
-        SQLiteCloud_number.value = 0
+        sqlitecloud_number = SQLiteCloudNumber()
+        sqlitecloud_number.value = 0
         extvalue = 0
-        isext = False
+        offcode = 0
+        isext = 0
         blen = len(buffer)
 
         # from 1 to skip the first command type character
         for i in range(index, blen):
             c = chr(buffer[i])
 
-            # check for optional extended error code (ERRCODE:EXTERRCODE)
+            # check for optional extended error code (ERRCODE:EXTERRCODE:OFFCODE)
             if c == ":":
-                isext = True
+                isext += 1
                 continue
 
             # check for end of value
             if c == " ":
-                SQLiteCloud_number.cstart = i + 1
-                SQLiteCloud_number.extcode = extvalue
-                return SQLiteCloud_number
+                sqlitecloud_number.cstart = i + 1
+                sqlitecloud_number.extcode = extvalue
+                sqlitecloud_number.offcode = offcode
+                return sqlitecloud_number
 
             val = int(c) if c.isdigit() else 0
 
-            # compute numeric value
-            if isext:
+            if isext == 1:
+                # XERRCODE
                 extvalue = (extvalue * 10) + val
+            elif isext == 2:
+                # OFFCODE
+                offcode = (offcode * 10) + val
             else:
-                SQLiteCloud_number.value = (SQLiteCloud_number.value * 10) + val
+                # generic value or ERRCODE
+                sqlitecloud_number.value = (sqlitecloud_number.value * 10) + val
 
-        SQLiteCloud_number.value = 0
-        return SQLiteCloud_number
+        sqlitecloud_number.value = 0
+        return sqlitecloud_number
 
     def _internal_parse_buffer(
         self, connection: SQLiteCloudConnect, buffer: bytes, blen: int
@@ -665,18 +678,16 @@ class Driver:
             if len_ == 0:
                 return SQLiteCloudResult(SQLITECLOUD_RESULT_TYPE.RESULT_STRING, "")
 
-            tag = (
-                SQLITECLOUD_RESULT_TYPE.RESULT_JSON
-                if cmd == SQLITECLOUD_CMD.JSON.value
-                else SQLITECLOUD_RESULT_TYPE.RESULT_STRING
-            )
+            tag = SQLITECLOUD_RESULT_TYPE.RESULT_STRING
 
             if cmd == SQLITECLOUD_CMD.ZEROSTRING.value:
                 len_ -= 1
             clone = buffer[cstart : cstart + len_]
 
             if cmd == SQLITECLOUD_CMD.COMMAND.value:
-                return self._internal_run_command(connection, clone)
+                return self._internal_run_command(
+                    connection, self._internal_serialize_command(clone.decode())
+                )
             elif cmd == SQLITECLOUD_CMD.PUBSUB.value:
                 return SQLiteCloudResult(
                     SQLITECLOUD_RESULT_TYPE.RESULT_OK,
@@ -693,12 +704,16 @@ class Driver:
                 )
             elif cmd == SQLITECLOUD_CMD.BLOB.value:
                 tag = SQLITECLOUD_RESULT_TYPE.RESULT_BLOB
+            elif cmd == SQLITECLOUD_CMD.JSON.value:
+                return SQLiteCloudResult(
+                    SQLITECLOUD_RESULT_TYPE.RESULT_JSON, json.loads(clone)
+                )
 
             clone = clone.decode() if cmd != SQLITECLOUD_CMD.BLOB.value else clone
             return SQLiteCloudResult(tag, clone)
 
         elif cmd == SQLITECLOUD_CMD.ERROR.value:
-            # -LEN ERRCODE:EXTCODE ERRMSG
+            # -LEN ERRCODE:EXTCODE:OFFCODE ERRMSG
             sqlite_number = self._internal_parse_number(buffer)
             len_ = sqlite_number.value
             cstart = sqlite_number.cstart
@@ -713,7 +728,9 @@ class Driver:
             len_ -= cstart2
             errmsg = clone[cstart2:]
 
-            raise SQLiteCloudException(errmsg.decode(), errcode, xerrcode)
+            raise get_sqlitecloud_error_with_extended_code(
+                errmsg.decode(), errcode, xerrcode
+            )(errmsg.decode(), errcode, xerrcode)
 
         elif cmd in [SQLITECLOUD_CMD.ROWSET.value, SQLITECLOUD_CMD.ROWSET_CHUNK.value]:
             # CMD_ROWSET:          *LEN 0:VERSION ROWS COLS DATA
@@ -740,11 +757,10 @@ class Driver:
                 rowset_signature.ncols,
             )
 
-            # continue parsing next chunk in the buffer
-            sign_len = rowset_signature.len
-            buffer = buffer[sign_len + len(f"/{sign_len} ") :]
-            if cmd == SQLITECLOUD_CMD.ROWSET_CHUNK.value and buffer:
-                return self._internal_parse_buffer(connection, buffer, len(buffer))
+            # continue reading from the socket
+            # until the end-of-chunk condition
+            if cmd == SQLITECLOUD_CMD.ROWSET_CHUNK.value:
+                return self._internal_socket_read(connection)
 
             return rowset
 
@@ -752,8 +768,8 @@ class Driver:
             return SQLiteCloudResult(SQLITECLOUD_RESULT_TYPE.RESULT_NONE, None)
 
         elif cmd in [SQLITECLOUD_CMD.INT.value, SQLITECLOUD_CMD.FLOAT.value]:
-            SQLiteCloud_value = self._internal_parse_value(buffer)
-            clone = SQLiteCloud_value.value
+            sqlitecloud_value = self._internal_parse_value(buffer)
+            clone = sqlitecloud_value.value
 
             tag = (
                 SQLITECLOUD_RESULT_TYPE.RESULT_INTEGER
@@ -784,6 +800,10 @@ class Driver:
         Returns:
             str: The uncompressed data.
         """
+        # buffer may contain a sequence of compressed data
+        # eg, a compressed rowset split in chunks is a sequence of rowset chunks
+        # compressed individually, each one with its compressed header,
+        # rowset header and compressed data
         space_index = buffer.index(b" ")
         buffer = buffer[space_index + 1 :]
 
@@ -821,14 +841,14 @@ class Driver:
 
         r: str = []
         for i in range(n):
-            SQLiteCloud_value = self._internal_parse_value(buffer, start)
-            start += SQLiteCloud_value.cellsize
-            r.append(SQLiteCloud_value.value)
+            sqlitecloud_value = self._internal_parse_value(buffer, start)
+            start += sqlitecloud_value.cellsize
+            r.append(sqlitecloud_value.value)
 
         return r
 
     def _internal_parse_value(self, buffer: bytes, index: int = 0) -> SQLiteCloudValue:
-        SQLiteCloud_value = SQLiteCloudValue()
+        sqlitecloud_value = SQLiteCloudValue()
         len = 0
         cellsize = 0
 
@@ -839,14 +859,15 @@ class Driver:
             if cellsize is not None:
                 cellsize = 2
 
-            SQLiteCloud_value.len = len
-            SQLiteCloud_value.cellsize = cellsize
+            sqlitecloud_value.value = None
+            sqlitecloud_value.len = len
+            sqlitecloud_value.cellsize = cellsize
 
-            return SQLiteCloud_value
+            return sqlitecloud_value
 
-        SQLiteCloud_number = self._internal_parse_number(buffer, index + 1)
-        blen = SQLiteCloud_number.value
-        cstart = SQLiteCloud_number.cstart
+        sqlitecloud_number = self._internal_parse_number(buffer, index + 1)
+        blen = sqlitecloud_number.value
+        cstart = sqlitecloud_number.cstart
 
         # handle decimal/float cases
         if c == SQLITECLOUD_CMD.INT.value or c == SQLITECLOUD_CMD.FLOAT.value:
@@ -854,20 +875,29 @@ class Driver:
             len = nlen - 2
             cellsize = nlen
 
-            SQLiteCloud_value.value = (buffer[index + 1 : index + 1 + len]).decode()
-            SQLiteCloud_value.len
-            SQLiteCloud_value.cellsize = cellsize
+            value = (buffer[index + 1 : index + 1 + len]).decode()
 
-            return SQLiteCloud_value
+            sqlitecloud_value.value = (
+                int(value) if c == SQLITECLOUD_CMD.INT.value else float(value)
+            )
+            sqlitecloud_value.len
+            sqlitecloud_value.cellsize = cellsize
+
+            return sqlitecloud_value
 
         len = blen - 1 if c == SQLITECLOUD_CMD.ZEROSTRING.value else blen
         cellsize = blen + cstart - index
 
-        SQLiteCloud_value.value = (buffer[cstart : cstart + len]).decode()
-        SQLiteCloud_value.len = len
-        SQLiteCloud_value.cellsize = cellsize
+        value = buffer[cstart : cstart + len]
 
-        return SQLiteCloud_value
+        if c == SQLITECLOUD_CMD.STRING.value or c == SQLITECLOUD_CMD.ZEROSTRING.value:
+            value = value.decode()
+
+        sqlitecloud_value.value = value
+        sqlitecloud_value.len = len
+        sqlitecloud_value.cellsize = cellsize
+
+        return sqlitecloud_value
 
     def _internal_parse_rowset_signature(
         self, buffer: bytes
@@ -951,9 +981,9 @@ class Driver:
         # parse column names
         rowset.colname = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            number_len = SQLiteCloud_number.value
-            cstart = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            number_len = sqlitecloud_number.value
+            cstart = sqlitecloud_number.cstart
             value = buffer[cstart : cstart + number_len]
             rowset.colname.append(value.decode())
             start = cstart + number_len
@@ -969,9 +999,9 @@ class Driver:
         # parse declared types
         rowset.decltype = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            number_len = SQLiteCloud_number.value
-            cstart = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            number_len = sqlitecloud_number.value
+            cstart = sqlitecloud_number.cstart
             value = buffer[cstart : cstart + number_len]
             rowset.decltype.append(value.decode())
             start = cstart + number_len
@@ -979,9 +1009,9 @@ class Driver:
         # parse database names
         rowset.dbname = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            number_len = SQLiteCloud_number.value
-            cstart = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            number_len = sqlitecloud_number.value
+            cstart = sqlitecloud_number.cstart
             value = buffer[cstart : cstart + number_len]
             rowset.dbname.append(value.decode())
             start = cstart + number_len
@@ -989,9 +1019,9 @@ class Driver:
         # parse table names
         rowset.tblname = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            number_len = SQLiteCloud_number.value
-            cstart = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            number_len = sqlitecloud_number.value
+            cstart = sqlitecloud_number.cstart
             value = buffer[cstart : cstart + number_len]
             rowset.tblname.append(value.decode())
             start = cstart + number_len
@@ -999,9 +1029,9 @@ class Driver:
         # parse column original names
         rowset.origname = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            number_len = SQLiteCloud_number.value
-            cstart = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            number_len = sqlitecloud_number.value
+            cstart = sqlitecloud_number.cstart
             value = buffer[cstart : cstart + number_len]
             rowset.origname.append(value.decode())
             start = cstart + number_len
@@ -1009,23 +1039,23 @@ class Driver:
         # parse not null flags
         rowset.notnull = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            rowset.notnull.append(SQLiteCloud_number.value)
-            start = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            rowset.notnull.append(sqlitecloud_number.value)
+            start = sqlitecloud_number.cstart
 
         # parse primary key flags
         rowset.prikey = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            rowset.prikey.append(SQLiteCloud_number.value)
-            start = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            rowset.prikey.append(sqlitecloud_number.value)
+            start = sqlitecloud_number.cstart
 
         # parse autoincrement flags
         rowset.autoinc = []
         for i in range(ncols):
-            SQLiteCloud_number = self._internal_parse_number(buffer, start)
-            rowset.autoinc.append(SQLiteCloud_number.value)
-            start = SQLiteCloud_number.cstart
+            sqlitecloud_number = self._internal_parse_number(buffer, start)
+            rowset.autoinc.append(sqlitecloud_number.value)
+            start = sqlitecloud_number.cstart
 
         return start
 
@@ -1034,6 +1064,52 @@ class Driver:
     ):
         # loop to parse each individual value
         for i in range(bound):
-            SQLiteCloud_value = self._internal_parse_value(buffer, start)
-            start += SQLiteCloud_value.cellsize
-            rowset.data.append(SQLiteCloud_value.value)
+            sqlitecloud_value = self._internal_parse_value(buffer, start)
+            start += sqlitecloud_value.cellsize
+            rowset.data.append(sqlitecloud_value.value)
+
+    def _internal_serialize_command(
+        self, data: Any, zero_string: bool = False
+    ) -> bytes:
+        if isinstance(data, str):
+            cmd = SQLITECLOUD_CMD.STRING.value
+            if zero_string:
+                cmd = SQLITECLOUD_CMD.ZEROSTRING.value
+                data += "\x00"
+
+            data = data.encode()
+            header = f"{cmd}{len(data)} ".encode()
+
+            return header + data
+
+        if isinstance(data, int):
+            return f"{SQLITECLOUD_CMD.INT.value}{data} ".encode()
+
+        if isinstance(data, float):
+            return f"{SQLITECLOUD_CMD.FLOAT.value}{data} ".encode()
+
+        if isinstance(data, bytes):
+            header = f"{SQLITECLOUD_CMD.BLOB.value}{len(data)} ".encode()
+            return header + data
+
+        if data is None:
+            return f"{SQLITECLOUD_CMD.NULL.value} ".encode()
+
+        if isinstance(data, List):
+            return self._internal_serialize_array(data, zero_string=zero_string)
+
+        raise SQLiteCloudException(
+            f"Unsupported data type for serialization: {type(data)}"
+        )
+
+    def _internal_serialize_array(self, data: List, zero_string: bool = False) -> str:
+        n = len(data)
+        serialized_data: bytes = f"{n} ".encode()
+        for i in range(n):
+            # the query must be zero-terminated
+            zs = i == 0 or zero_string
+            serialized_data += self._internal_serialize_command(data[i], zero_string=zs)
+
+        header = f"{SQLITECLOUD_CMD.ARRAY.value}{len(serialized_data)} ".encode()
+
+        return header + serialized_data
