@@ -1,6 +1,26 @@
+from enum import Enum
 from typing import Any, Dict, List, Optional
 
-from sqlitecloud.types import SQLITECLOUD_RESULT_TYPE, SQLITECLOUD_VALUE_TYPE
+
+class SQLITECLOUD_VALUE_TYPE(Enum):
+    INTEGER = "INTEGER"
+    FLOAT = "REAL"
+    TEXT = "TEXT"
+    BLOB = "BLOB"
+    NULL = "NULL"
+
+
+class SQLITECLOUD_RESULT_TYPE(Enum):
+    RESULT_OK = 0
+    RESULT_ERROR = 1
+    RESULT_STRING = 2
+    RESULT_INTEGER = 3
+    RESULT_FLOAT = 4
+    RESULT_ROWSET = 5
+    RESULT_ARRAY = 6
+    RESULT_NONE = 7
+    RESULT_JSON = 8
+    RESULT_BLOB = 9
 
 
 class SQLiteCloudResult:
@@ -40,35 +60,23 @@ class SQLiteCloudResult:
             return -1
         return row * self.ncols + col
 
-    def get_value(self, row: int, col: int, convert: bool = True) -> Optional[any]:
+    def get_value(self, row: int, col: int) -> Optional[any]:
         index = self._compute_index(row, col)
         if index < 0 or not self.data or index >= len(self.data):
             return None
 
-        value = self.data[index]
-        return self._convert(value, col) if convert else value
+        return self.data[index]
 
     def get_name(self, col: int) -> Optional[str]:
         if col < 0 or col >= self.ncols:
             return None
         return self.colname[col]
 
-    def _convert(self, value: str, col: int) -> any:
-        if col < 0 or col >= len(self.decltype):
-            return value
-
-        decltype = self.decltype[col]
-        if decltype == SQLITECLOUD_VALUE_TYPE.INTEGER.value:
-            return int(value)
-        if decltype == SQLITECLOUD_VALUE_TYPE.FLOAT.value:
-            return float(value)
-        if decltype == SQLITECLOUD_VALUE_TYPE.BLOB.value:
-            # values are received as bytes before being strings
-            return bytes(value)
-        if decltype == SQLITECLOUD_VALUE_TYPE.NULL.value:
+    def get_decltype(self, col: int) -> Optional[str]:
+        if col < 0 or col >= self.ncols or col >= len(self.decltype):
             return None
 
-        return value
+        return self.decltype[col]
 
 
 class SQLiteCloudResultSet:
@@ -106,3 +114,34 @@ class SQLiteCloudResultSet:
 
     def get_result(self) -> Optional[any]:
         return self.get_value(0, 0)
+
+
+class SQLiteCloudOperationResult:
+    """Result of a DML operation in a SQLite statement."""
+
+    def __init__(self, result: SQLiteCloudResult) -> None:
+        self._result = result
+
+    @property
+    def type(self) -> int:
+        return self._result.data[0][0]
+
+    @property
+    def index(self) -> int:
+        return self._result.data[0][1]
+
+    @property
+    def rowid(self) -> int:
+        return self._result.data[0][2]
+
+    @property
+    def changes(self) -> int:
+        return self._result.data[0][3]
+
+    @property
+    def total_changes(self) -> int:
+        return self._result.data[0][4]
+
+    @property
+    def finalized(self) -> int:
+        return self._result.data[0][5]
