@@ -1,18 +1,22 @@
-import json
 import os
+import random
 import time
 
 import pytest
 
 from sqlitecloud.client import SQLiteCloudClient
-from sqlitecloud.types import (
+from sqlitecloud.datatypes import (
     SQLITECLOUD_ERRCODE,
     SQLITECLOUD_INTERNAL_ERRCODE,
-    SQLITECLOUD_RESULT_TYPE,
     SQLiteCloudAccount,
     SQLiteCloudConnect,
-    SQLiteCloudException,
 )
+from sqlitecloud.exceptions import (
+    SQLiteCloudError,
+    SQLiteCloudException,
+    SQLiteCloudOperationalError,
+)
+from sqlitecloud.resultset import SQLITECLOUD_RESULT_TYPE
 
 
 class TestClient:
@@ -56,7 +60,7 @@ class TestClient:
 
         client = SQLiteCloudClient(cloud_account=account)
 
-        with pytest.raises(SQLiteCloudException):
+        with pytest.raises(SQLiteCloudError):
             client.open_connection()
 
     def test_connect_with_string(self):
@@ -124,7 +128,7 @@ class TestClient:
 
     def test_column_not_found(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
-        with pytest.raises(SQLiteCloudException) as e:
+        with pytest.raises(SQLiteCloudOperationalError) as e:
             client.exec_query("SELECT not_a_column FROM albums", connection)
 
         assert e.value.errcode == 1
@@ -248,7 +252,7 @@ class TestClient:
                 "soldier: Well they don't seem to move anymore...",
                 "supreme-commander: Oh snap, I came here to see them twerk!",
             ],
-        } == json.loads(result.get_result())
+        } == result.get_result()
 
     def test_blob(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
@@ -257,7 +261,7 @@ class TestClient:
         assert SQLITECLOUD_RESULT_TYPE.RESULT_BLOB == result.tag
         assert len(result.get_result()) == 1000
 
-    def test_blob0(self, sqlitecloud_connection):
+    def test_blob_zero_length(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
         result = client.exec_query("TEST BLOB0", connection)
 
@@ -267,7 +271,7 @@ class TestClient:
     def test_error(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
 
-        with pytest.raises(SQLiteCloudException) as e:
+        with pytest.raises(SQLiteCloudError) as e:
             client.exec_query("TEST ERROR", connection)
 
         assert e.value.errcode == 66666
@@ -276,7 +280,7 @@ class TestClient:
     def test_ext_error(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
 
-        with pytest.raises(SQLiteCloudException) as e:
+        with pytest.raises(SQLiteCloudError) as e:
             client.exec_query("TEST EXTERROR", connection)
 
         assert e.value.errcode == 66666
@@ -296,8 +300,8 @@ class TestClient:
         assert isinstance(result_array, list)
         assert len(result_array) == 5
         assert result_array[0] == "Hello World"
-        assert result_array[1] == "123456"
-        assert result_array[2] == "3.1415"
+        assert result_array[1] == 123456
+        assert result_array[2] == 3.1415
         assert result_array[3] is None
 
     def test_rowset(self, sqlitecloud_connection):
@@ -311,6 +315,19 @@ class TestClient:
         assert result.get_name(0) == "key"
         assert result.get_name(1) == "value"
 
+    def test_rowset_data_types(self, sqlitecloud_connection):
+        connection, client = sqlitecloud_connection
+
+        bindings = ("hello world", 15175, 3.14, b"bytes world", None)
+        result = client.exec_statement("SELECT ?, ?, ?, ?, ?", bindings, connection)
+
+        assert SQLITECLOUD_RESULT_TYPE.RESULT_ROWSET == result.tag
+        assert result.get_value(0, 0) == "hello world"
+        assert result.get_value(0, 1) == 15175
+        assert result.get_value(0, 2) == 3.14
+        assert result.get_value(0, 3) == b"bytes world"
+        assert result.get_value(0, 4) is None
+
     def test_max_rows_option(self):
         account = SQLiteCloudAccount()
         account.hostname = os.getenv("SQLITE_HOST")
@@ -341,7 +358,7 @@ class TestClient:
 
         connection = client.open_connection()
 
-        with pytest.raises(SQLiteCloudException) as e:
+        with pytest.raises(SQLiteCloudError) as e:
             client.exec_query("SELECT * FROM albums", connection)
 
         client.disconnect(connection)
@@ -366,7 +383,7 @@ class TestClient:
 
         assert 1 == rowset.nrows
 
-    def test_chunked_rowset(self, sqlitecloud_connection):
+    def test_rowset_chunk(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
 
         rowset = client.exec_query("TEST ROWSET_CHUNK", connection)
@@ -377,6 +394,17 @@ class TestClient:
         assert 147 == len(rowset.data)
         assert "key" == rowset.get_name(0)
 
+    def test_rowset_nochunk(self, sqlitecloud_connection):
+        connection, client = sqlitecloud_connection
+
+        rowset = client.exec_query("TEST ROWSET_NOCHUNK", connection)
+
+        assert SQLITECLOUD_RESULT_TYPE.RESULT_ROWSET == rowset.tag
+        assert 147 == rowset.nrows
+        assert 1 == rowset.ncols
+        assert 147 == len(rowset.data)
+        assert "key" == rowset.get_name(0)
+
     def test_chunked_rowset_twice(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
         rowset = client.exec_query("TEST ROWSET_CHUNK", connection)
@@ -409,7 +437,7 @@ class TestClient:
             assert 2 == rowset.ncols
             assert "count" == rowset.get_name(0)
             assert "string" == rowset.get_name(1)
-            assert str(i) == rowset.get_value(0, 0)
+            assert i == rowset.get_value(0, 0)
             assert rowset.version in [1, 2]
 
     def test_query_timeout(self):
@@ -441,7 +469,10 @@ class TestClient:
         client.disconnect(connection)
 
         assert e.value.errcode == SQLITECLOUD_INTERNAL_ERRCODE.NETWORK
-        assert e.value.errmsg == "An error occurred while reading data from the socket."
+        assert (
+            e.value.errmsg
+            == "An error occurred while reading command length from the socket."
+        )
 
     def test_XXL_query(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
@@ -491,7 +522,7 @@ class TestClient:
         assert rowset.ncols == 2
         assert rowset.get_name(0) == "42"
         assert rowset.get_name(1) == "'hello'"
-        assert rowset.get_value(0, 0) == "42"
+        assert rowset.get_value(0, 0) == 42
         assert rowset.get_value(0, 1) == "hello"
 
     def test_select_long_formatted_string(self, sqlitecloud_connection):
@@ -603,6 +634,43 @@ class TestClient:
                         query_ms < self.EXPECT_SPEED_MS
                     ), f"{num_queries}x batched selects, {query_ms}ms per query"
 
+    @pytest.mark.slow
+    def test_big_rowset(self):
+        account = SQLiteCloudAccount()
+        account.hostname = os.getenv("SQLITE_HOST")
+        account.apikey = os.getenv("SQLITE_API_KEY")
+        account.dbname = os.getenv("SQLITE_DB")
+
+        client = SQLiteCloudClient(cloud_account=account)
+        client.config.timeout = 120
+
+        connection = client.open_connection()
+
+        table_name = "TestCompress" + str(random.randint(0, 99999))
+        try:
+            client.exec_query(
+                f"CREATE TABLE {table_name} (id INTEGER PRIMARY KEY, name TEXT)",
+                connection,
+            )
+
+            nRows = 5000
+
+            sql = ""
+            for i in range(nRows):
+                sql += f"INSERT INTO {table_name} (name) VALUES ('Test-{i}'); "
+
+            client.exec_query(sql, connection)
+
+            rowset = client.exec_query(
+                f"SELECT * from {table_name}",
+                connection,
+            )
+
+            assert rowset.nrows == nRows
+        finally:
+            client.exec_query(f"DROP TABLE IF EXISTS {table_name}", connection)
+            client.disconnect(connection)
+
     def test_compression_single_column(self):
         account = SQLiteCloudAccount()
         account.hostname = os.getenv("SQLITE_HOST")
@@ -651,17 +719,27 @@ class TestClient:
         assert rowset.ncols > 0
         assert rowset.get_name(0) == "AlbumId"
 
-    def test_exec_statement_with_named_placeholder(self, sqlitecloud_connection):
+    def test_rowset_nochunk_compressed(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
 
-        result = client.exec_statement(
-            "SELECT * FROM albums WHERE AlbumId = :id and Title = :title",
-            {"id": 1, "title": "For Those About To Rock We Salute You"},
-            connection,
-        )
+        rowset = client.exec_query("TEST ROWSET_NOCHUNK_COMPRESSED", connection)
 
-        assert result.nrows == 1
-        assert result.get_value(0, 0) == 1
+        assert SQLITECLOUD_RESULT_TYPE.RESULT_ROWSET == rowset.tag
+        assert 147 == rowset.nrows
+        assert 1 == rowset.ncols
+        assert 147 == len(rowset.data)
+        assert "key" == rowset.get_name(0)
+
+    def test_rowset_chunk_compressed(self, sqlitecloud_connection):
+        connection, client = sqlitecloud_connection
+
+        rowset = client.exec_query("TEST ROWSET_CHUNK_COMPRESSED", connection)
+
+        assert SQLITECLOUD_RESULT_TYPE.RESULT_ROWSET == rowset.tag
+        assert 147 == rowset.nrows
+        assert 1 == rowset.ncols
+        assert 147 == len(rowset.data)
+        assert "key" == rowset.get_name(0)
 
     def test_exec_statement_with_qmarks(self, sqlitecloud_connection):
         connection, client = sqlitecloud_connection
